
#include <sys/ioctl.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

#include <linux/dvb/dmx.h>
#include <linux/dvb/frontend.h>

#include <event.h>

#include "getstream.h"
#include "stream.h"
#include "crc32.h"

static uint8_t			dvrbuf[DVR_BUFFER_SIZE];
static unsigned long long	packets;

static struct pidentry_s pidtable[MAX_PID];

/*
 * TS Header
 *
 * Name				Bits
 *
 * sync_byte			 8   (0x47)
 * transport_error_indicator	 1
 * payload_unit_start_indicator	 1
 * transport_priority		 1
 * PID				13
 * transport_scrambling_control	 2
 * adaption_field_control	 2
 * continuity_counter		 4
 *
 * if (adaption_field_control == '10' ||
 *	adaption_field_control == '11') {
 *
 *	adaption_field();
 * }
 *
 */


/*
 * DMX_PES_VIDEO
 * DMX_PES_AUDIO
 * DMX_PES_TELETEXT
 * DMX_PES_OTHER
 *
 */
int demux_set_pes_filter(int fd, int pid, int pestype) {
	struct dmx_pes_filter_params pesFilterParams;

	fprintf(stderr,"Setting filter for pid %d pestype %d\n", pid, pestype);

	pesFilterParams.pid		= pid;
	pesFilterParams.input		= DMX_IN_FRONTEND;
	pesFilterParams.output		= DMX_OUT_TS_TAP;
	pesFilterParams.pes_type	= pestype;
	pesFilterParams.flags		= DMX_IMMEDIATE_START;

	if (ioctl(fd, DMX_SET_PES_FILTER, &pesFilterParams) < 0)  {
		fprintf(stderr,"ioctl DMX_SET_PES_FILTER failed for pid %u pestype %d ",pid, pestype);
		exit(-1);
	}

	return 0;
}

/*
 * Apply section filter to opened demux interface.
 *
 */
int demux_set_sct_filter(int fd, int pid,
		struct dmx_filter *df, int flags, int timeout) {

	struct dmx_sct_filter_params	sctFilterParams;

	memset(&sctFilterParams, 0, sizeof(struct dmx_sct_filter_params));

	sctFilterParams.pid=pid;
	sctFilterParams.timeout=timeout;
	sctFilterParams.flags=flags;

	memcpy(&sctFilterParams.filter, df, sizeof(struct dmx_filter));

	printf("filter %02x\n", sctFilterParams.filter.filter[0]);

	if (ioctl(fd, DMX_SET_FILTER, &sctFilterParams) < 0) {
		fprintf(stderr,"ioctl DMX_SET_PES_FILTER failed for pid %u\n",pid);
		exit(-1);
	}

	return 0;
}

/*
 * 0000 47 4d 48 1d 00 02 b0 2f 6d dc df 00 00 ed 49 f0
 *
 *      aa bb bb cd
 *      a - TS Sync
 *      b - TS PID (13 bit) + payload_unit_start_indicator
 *      c - TS scrambling / adaption
 *      d - TS continuity counter
 *
 * 0010 00 04 ed 49 f0 00 05 e8 16 f0 05 6f 03 00 01 e0
 * 0020 0b e8 18 f0 0e 13 05 00 00 00 00 00 52 01 0a 66
 * 0030 02 00 f0 88 ab c1 a5 ff ff ff ff ff ff ff ff ff
 * 0040 ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff
 * 0050 ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff
 * 0060 ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff
 * 0070 ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff
 * 0080 ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff
 * 0090 ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff
 * 00a0 ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff
 * 00b0 ff ff ff ff ff ff ff ff ff ff ff ff
 *
 */
#define TS_SYNC_OFF	0
#define TS_SYNC		0x47

#define TS_PID_OFF1	1
#define TS_PID_OFF2	2

#define TS_PUSI_OFF	1
#define TS_PUSI_MASK	0x40

#define TS_AFC_OFF	3
#define TS_AFC_MASK	0x30
#define TS_AFC_SHIFT	4

#define TS_HEAD_MIN	4
#define TS_AFC_LEN	4

#define TS_CC_OFF	3
#define TS_CC_MASK	0xf

#define TS_AF_OFF	4
#define TS_PAYLOAD_OFF	4

int ts_get_payload_off(uint8_t *tsp, int psi) {
	int off=TS_HEAD_MIN,
	    pusi,
	    afc;

	if (tsp[TS_SYNC_OFF] != TS_SYNC)
		return 0;

	pusi=tsp[TS_PUSI_OFF] && TS_PUSI_MASK;

	/*
	 * If payload_unit_start_indicator == 1 and
	 * we have a PSI stream we have a pointer to skip
	 */
	if (pusi && psi) {
		off++;
		off+=tsp[TS_PAYLOAD_OFF];
	}

	afc=(tsp[TS_AFC_OFF] & TS_AFC_MASK) >> TS_AFC_SHIFT;

	/*
	 * AFC
	 * %00	-> Reserved
	 * %01	-> No Adaption field, payload only
	 * %10	-> Adaption field only, no payload
	 * %11	-> Adaption field followed by payload
	 */
	if (afc & 0x2)
		off+=tsp[TS_AFC_LEN];

	return off;
}


/*
 * PMT - Program Map Table 
 *
 * Name				No of Bits	Offset
 * table_id			 8		 0
 * section_syntax_indicator	 1		 1
 * '0'				 1		 1
 * reserved			 2		 1
 * section_length		12		 1/2
 * programm_number		16		 3/4
 * reserved			 2		 5
 * version_number		 5		 5
 * current_next_indicator	 1		 5
 * section_number		 8		 6
 * last_section_number		 8		 7
 * reserved			 3		 8
 * PCR_PID			13		 8/9
 * reserved			 4		10
 * programm_info_length		12		10/11
 * 1 ... N (program_info_length)
 *  descriptor()
 *
 * 1 ... N1
 *  stream_type			 8		0
 *  reserved			 3		1
 *  elementary_PID		13		1/2
 *  reserved			 4		3
 *  ES_info_length		12		3/4
 *  1 ... N2
 *   descriptor()
 *
 * CRC32			32
 */

#define PMT_TABLE_OFF		0
#define PMT_TABLE_ID		0x02
#define PMT_SECLEN_OFF1		1
#define PMT_SECLEN_OFF2		2
#define PMT_SECLEN_MASK		0x0fff
#define PMT_PNR_OFF1		3
#define PMT_PNR_OFF2		4
#define PMT_PILEN_OFF1		10
#define PMT_PILEN_OFF2		11
#define PMT_PILEN_MASK		0x03ff
#define PMT_PI_OFF		12

#define PMT_ST_STYPE_OFF		0
#define PMT_ST_PID_OFF1		1
#define PMT_ST_PID_OFF2		2

#define PMT_ST_ESLEN_OFF1	3
#define PMT_ST_ESLEN_OFF2	4

#define PMT_ST_ESLEN_MASK	0x0fff
#define PMT_ST_ES_OFF		5

#define PMT_PID_MASK		0x1fff

#define PMT_D_TAG_OFF		0
#define PMT_D_LEN_OFF		1

#define PMT_SECTION_OFF		5
#define PMT_LAST_SECTION_OFF	5

#define PMT_CRC32_LEN		4

static void parse_pmt(struct channel_s *channel, uint8_t *ts) {
	int	espid, eslen, pilen,  pid,
		pmtseclen, pnr, tspl, sto, ste,
		pmtmaxseclen, pmtlen;
	time_t	now;
	uint8_t	*pmt;

	/* Get offset to TS (Transport Stream) payload. The PMT
	 * is a PSI (Program Specific Information) so we dont
	 * have a PES header */
	tspl=ts_get_payload_off(ts, 1);

	pmt=&ts[tspl];

	pid=(ts[TS_PID_OFF1]<<8|ts[TS_PID_OFF2])&PMT_PID_MASK;

	/*
	 * Is this really a PMT ? It is actually a very common
	 * thing to have section with table_id != PMT on the PMT
	 * pid. Silently ignore the packets
	 *
	 */
	if (pmt[PMT_TABLE_OFF] != PMT_TABLE_ID) {
#if 0
		fprintf(stderr, "TS packet pid %d not a PMT packet table_id %02x\n",
				pid, pmt[PMT_TABLE_OFF]);
#endif
		return;
	}

	/* Length of PMT */
	pmtseclen=(pmt[PMT_SECLEN_OFF1] << 8 |
		pmt[PMT_SECLEN_OFF2]) & PMT_SECLEN_MASK;

	/* Maximum PMT Payload
	 *
	 * TS_PACKET_SIZE minus
	 * - TS header (tspl)
	 * - starts behind section_length
	 *
	 */

	pmtmaxseclen=TS_PACKET_SIZE-tspl-PMT_SECLEN_OFF2;

	/* Check whether we are overflowing the TS packet
	 *
	 * This is actually a common error found on different
	 * transponders. Silently use MIN.
	 */
	pmtlen=(pmtseclen > pmtmaxseclen) ?
		pmtmaxseclen : pmtseclen;

	/* Get program number */
	pnr=pmt[PMT_PNR_OFF1] << 8 | pmt[PMT_PNR_OFF2];

	/* Get program descriptor length */
	pilen=(pmt[PMT_PILEN_OFF1] << 8 | pmt[PMT_PILEN_OFF2]) & PMT_PILEN_MASK;

	now=time(NULL);

	if (pnr != channel->id) {
		fprintf(stderr, "Received PMT on pid %d - not for pnr %d\n",
				pid, pnr);
		return;
	}

	/* Start of Stream Table is header + programm_info descriptors */
	sto=PMT_PI_OFF+pilen;
	/* End of Stream Table */
	ste=pmtlen+PMT_SECLEN_OFF2-PMT_CRC32_LEN;

	while(sto < ste) {
		espid=(pmt[sto+PMT_ST_PID_OFF1]<<8 | pmt[sto+PMT_ST_PID_OFF2]) & PMT_PID_MASK;
		eslen=(pmt[sto+PMT_ST_ESLEN_OFF1]<<8 | pmt[sto+PMT_ST_ESLEN_OFF2]) & PMT_ST_ESLEN_MASK;

		switch (pmt[sto]) {
			case 1:		/* ISO/IEC 11172 Video		*/

			case 2:		/* ITU-T Rec. H.262		*/
					/* ISO/IEC 13818-2 Video	*/
					/* ISO/IEC 11172-2		*/
				if (pidtable[espid].type == PID_NONE) {
					pidtable[espid].type = PID_VIDEO;
					pidtable[espid].channel = channel;
					printf("Setting PID %d type to VIDEO for PNR %04x (%s)\n",
						espid, channel->id,
						channel->name);
					channel->vpid++;
				}
				pidtable[espid].lastvalid = now;
				break;

			case 3:		/* ISO/IEC 11172 Audio */

			case 4:		/* ISO/IEC 13818-3 Audio */

				if (pidtable[espid].type == PID_NONE) {
					pidtable[espid].type = PID_AUDIO;
					pidtable[espid].channel = channel;
					printf("Setting PID %d type to AUDIO for PNR %04x (%s)\n",
						espid, channel->id,
						channel->name);
					channel->apid++;
				}
				pidtable[espid].lastvalid = now;
				break;

			case 6:		/* ITU-T Rec. H.222.0 */
					/* ISO/IEC 13818-1 PES */

				if (pidtable[espid].type == PID_NONE) {
					pidtable[espid].type = PID_OTHER;
					pidtable[espid].channel = channel;
					printf("Setting PID %d type to OTHER for PNR %04x (%s)\n",
						espid, channel->id,
						channel->name);
					channel->opid++;
				}
				pidtable[espid].lastvalid = now;
				break;
			default:
				/* printf("pid %d as unknwown type %d\n", espid, pmt[sto]); */
				break;
			}
#if 0
#warn This code is not used anywhere

			/* Parse ES "Elemetary Stream" descriptors */
			if (eslen) {
				int esdo=sto+PMT_ST_ES_OFF;	/* ES descriptor offset */
				int esde=esdo+eslen;		/* ES descriptor end */
				int dlen;

				/* Loop on ES descriptors */
				do {
					dlen=pmt[esdo+PMT_D_LEN_OFF] + PMT_D_LEN_OFF + 1;

					switch(pmt[esdo+PMT_D_TAG_OFF]) {
						case 10:	/* ISO 639 language descriptor */
						/* FIXME - Language is important */
						break;
				}

				/* Oops !?!? */
				if (!dlen) {
					printf("ES descriptor length went 0\n");
					break;
				}

				esdo+=dlen;
			/* Abort if we are beyond end */
			} while(esdo<esde);

		}
#endif
		/* Add Stream table descriptor entry size */
		sto+=PMT_ST_ES_OFF+eslen;
	}
}

static void dvr_read(int fd, short event, void *arg) {
	int	len, i, cc;

	do {
		len=read(fd, dvrbuf, DVR_BUFFER_SIZE);

		if (len < 0)
			break;

		/* We should only get n*TS_PACKET_SIZE */
		if (len % TS_PACKET_SIZE != 0) {
			printf("Oops - unaligned read of %d bytes dropping buffer\n", len);
			continue;
		}

		packets+=len/TS_PACKET_SIZE;

		for(i=0;i<len;i+=TS_PACKET_SIZE) {

			/* Continuity counter */
			cc=dvrbuf[i+TS_CC_OFF] & TS_CC_MASK;

			/* TS (Transport Stream) packets start with 0x47 */
			if (dvrbuf[i] != 0x47) {
#ifdef DEBUG
				printf("Non TS Stream packet (!0x47) received on dvr0\n");
				dump_hex(&dvrbuf[i], TS_PACKET_SIZE);
#endif
				continue;
			}

#define TS_PID_MASK	0x1fff
			int pid=(dvrbuf[i+1]<<8|dvrbuf[i+2]) & TS_PID_MASK;

			/* Does somebody want this pid ? */
			if (!pidtable[pid].channel)
				continue;
#if 0
			/* FIXME
			 * - Retransmission of the same packet may only happen on two
			 *   consecutive packets
			 * - If the AFC was field is %00 || %01 the cc will not be increased
			 * - Initial discontinuity should be silently suppressed.
			 */
			if (((pidtable[pid].cc+1)&TS_CC_MASK) != cc) {
				printf("Discontinuity in pid %d %d vs %d (i=%d, len=%d)\n",
						pid, pidtable[pid].cc, cc, i, len);
			}
			pidtable[pid].cc=cc;
#endif

			switch(pidtable[pid].type) {
				case(PID_VIDEO):
				case(PID_AUDIO):
				case(PID_OTHER):
					stream_send(pidtable[pid].channel, &dvrbuf[i]);
					break;
				case(PID_PMT):
					parse_pmt(pidtable[pid].channel, &dvrbuf[i]);
					stream_send(pidtable[pid].channel, &dvrbuf[i]);
					break;
			}
		}
	} while (len > 0);
}


/*
 * PAT Programm Association Table
 *
 * The PAT is multiplexed into the TS (Transport Stream)
 * at PID 0x0. It contains the table of all PMT (Program Map Table)
 * pids within the TS.
 *
 * Name				Bits	Offset
 *
 * table_id			 8	 0
 * section_syntax_indicator	 1	 1
 * pad (0)			 1	 1
 * reserved			 2	 1
 * section_length		12	 1/2
 * transport_stream_id		16	 3/4
 * reserved			 2	 5
 * version_number		 5	 5
 * current_next_indicator	 1	 5
 * section_number		 8	 6
 * last_section_number		 8	 7
 *
 * 1..N
 *   program_number		16
 *   reserved			 3
 *   pid			13
 *      (if pnum==0 network_pid else program_map_pid)
 *
 * crc32			32
 *
 */

#define PAT_TABLE_ID_OFF	0
#define PAT_TABLE_ID		0x0

#define PAT_LAST_SECTION_OFF	7
#define PAT_SECTION_OFF		6

#define PAT_SLEN_OFF1		1
#define PAT_SLEN_OFF2		2
#define PAT_SLEN_MASK		0x0fff

#define PAT_TID_OFF1		3
#define PAT_TID_OFF2		4

#define PAT_VER_OFF		5

#define PAT_PNR_OFF1		8	/* First PNR offset byte 1 */
#define PAT_PNR_OFF2		9	/* First PNR offset byte 2 */

#define PAT_PMTPID_OFF1		10	/* First PMTPID offset byte 1 */
#define PAT_PMTPID_OFF2		11	/* First PMTPID offset byte 2 */
#define PAT_PMTPID_MASK		0x1fff

#define PAT_CRC_OFF1		12	/* Offset of CRC32 in case of single program PAT */
#define PAT_CRC_OFF2		13
#define PAT_CRC_OFF3		14
#define PAT_CRC_OFF4		15

#define PAT_MIN_LEN		9

/*
 * 0000 47 40 64 18 00 02 b0 83 6d ca f5 00 00 e0 65 f0    G.d.....m.....e.
 *      aa bc cc de
 *
 * a - TS_SYNC					-> 0x47
 * b -	transport_error_indicator	0
 *	payload_unit_start_indicator	1
 *	transport_priority		0	-> 0x4
 * c - PID					-> 0x64
 * d -	transport_scrambling_control	00
 *	adaption_field_control		01	-> 0x1
 * e - continuity_counter			-> 0x8
 *
 */
static void dmx_send_pat(uint8_t *rpat, struct adapter_s *adapter) {
	uint8_t			tsbuf[TS_PACKET_SIZE],
				*pat;
	GList			*cl;
	unsigned int		sid;
	__u32			crc;

	/* Fill with 0xff */
	memset(tsbuf, 0xff, TS_PACKET_SIZE);

	cl=g_list_first(adapter->channel);
	while(cl) {
		struct channel_s	*channel=cl->data;

		channel->pid0cc=(channel->pid0cc+1)&TS_CC_MASK;	/* Increment CC */

		sid=channel->id;

		/* Build PID 0 TS packet */
		tsbuf[TS_SYNC_OFF]=TS_SYNC;
		tsbuf[TS_PID_OFF1]=0x0;			/* Clear PID */
		tsbuf[TS_PID_OFF2]=0x0;
		tsbuf[TS_PUSI_OFF]=TS_PUSI_MASK;	/* PUSI == 0x40 */
		tsbuf[TS_CC_OFF]=(channel->pid0cc&TS_CC_MASK);/* Store CC into TS */
		tsbuf[TS_AFC_OFF]|=0x1<<TS_AFC_SHIFT;	/* Payload only - no adaption field */
		tsbuf[TS_PAYLOAD_OFF]=0x0;		/* Clear PSI pointer */

		pat=&tsbuf[TS_PAYLOAD_OFF+1];

		pat[PAT_TABLE_ID_OFF]=PAT_TABLE_ID;
		pat[PAT_SLEN_OFF1]=0x0 | 0x80;		/* Set section_syntax_inidicator */
		pat[PAT_SLEN_OFF2]=13;			/* Single Programm PAT */
		pat[PAT_TID_OFF1]=rpat[PAT_TID_OFF1];
		pat[PAT_TID_OFF2]=rpat[PAT_TID_OFF2];
		pat[PAT_VER_OFF]=rpat[PAT_VER_OFF];
		pat[PAT_SECTION_OFF]=0x0;
		pat[PAT_LAST_SECTION_OFF]=0x0;

		pat[PAT_PNR_OFF1]=(sid >> 8) & 0xff;
		pat[PAT_PNR_OFF2]=sid & 0xff;
		pat[PAT_PMTPID_OFF1]=(channel->ppid >> 8) & 0xff;
		pat[PAT_PMTPID_OFF2]=(channel->ppid) & 0xff;

		pat[PAT_CRC_OFF1]=0x0;
		pat[PAT_CRC_OFF2]=0x0;
		pat[PAT_CRC_OFF3]=0x0;
		pat[PAT_CRC_OFF4]=0x0;

		/* CRC over complete pat without CRC */
		crc=calc_crc32(pat, PAT_CRC_OFF1);

		pat[PAT_CRC_OFF1]=crc >> 24 & 0xff;
		pat[PAT_CRC_OFF2]=crc >> 16 & 0xff;
		pat[PAT_CRC_OFF3]=crc >> 8 & 0xff;
		pat[PAT_CRC_OFF4]=crc & 0xff;

		stream_send(channel, tsbuf);

		cl=g_list_next(cl);
	}


/*
 * TS Header
 *
 * Name				Bits
 *
 * sync_byte			 8   (0x47)
 * transport_error_indicator	 1
 * payload_unit_start_indicator	 1
 * transport_priority		 1
 * PID				13
 * transport_scrambling_control	 2
 * adaption_field_control	 2
 * continuity_counter		 4
 *
 * if (adaption_field_control == '10' ||
 *	adaption_field_control == '11') {
 *
 *	adaption_field();
 * }
 *
 */

}


/*
 * Parse PAT (Programm Association Table) and set pid types
 * in the pid demux table.
 */
static void dmx_parse_pat(uint8_t *buf, int len,
		struct adapter_s *adapter) {
	int			ppid, pnr, patlen, patnoppid, i;
	time_t			now;
	GList			*cl;

	now=time(NULL);

	if (len < PAT_MIN_LEN)
		return;

	/*
	 * FIXME - We shoudl check for current_next bit
	 * to make shure we only parse and process
	 * current PATs
	 *
	 */

	/* Calculate this PATs length */
	patlen=(buf[PAT_SLEN_OFF1] << 8 | buf[PAT_SLEN_OFF2])
		& PAT_SLEN_MASK;

	/* Calculate the number PMTpids in this PAT */
	patnoppid=(patlen-9) / 4;

	/* Loop on ppids */
	for(i=0;i<patnoppid;i++) {
		/* Get Programm Number */
		pnr=buf[PAT_PNR_OFF1+i*4] << 8 | buf[PAT_PNR_OFF2+i*4];
		ppid=(buf[PAT_PMTPID_OFF1+i*4] << 8 | buf[PAT_PMTPID_OFF2+i*4])
				& PAT_PMTPID_MASK;

		/*
		 * Try to check whether we want this program number
		 * We do a linear search here as there may only be up
		 * to 30-40 Programs per transponder and PATs are "rare"
		 *
		 */
		cl=g_list_first(adapter->channel);
		while(cl) {
			struct channel_s	*channel=cl->data;

			if (channel->id == pnr) {
				/*
				 * If we do want this program - Feed the PMTs PID into
				 * the pid demux table
				 */
				channel->ppid=ppid;

				if (pidtable[ppid].type != PID_PMT) {
					printf("Setting PID %d type to PMT for PNR %04x\n",
						ppid, channel->id);
					channel->opid++;
					pidtable[ppid].type = PID_PMT;
					pidtable[ppid].channel = channel;
				}

				/* Set PID to type PMT and store stream ptr */
				pidtable[ppid].lastvalid = now;

				break;
			}

			cl=g_list_next(cl);
		}
	}
}


/*
 * This will be called for PAT (Programm Association Table)
 * receive on PID 0. We should parse the PAT and fill the pid
 * demux table with the PMT pids pointing to the PMT decoder.
 *
 */
#define MAX_SEC_LEN	8192
static void dmx_read(int fd, short event, void *arg) {
	struct adapter_s	*adapter=arg;
	static uint8_t		secbuf[MAX_SEC_LEN];
	int			len;

	len=read(fd, &secbuf, MAX_SEC_LEN);

	if (len) {
		dmx_parse_pat(secbuf, len, adapter);
		dmx_send_pat(secbuf, adapter);
	}
}

static inline char *dvrname(int adapter) {
	static char	dvrname[128];

	sprintf(dvrname, "/dev/dvb/adapter%d/dvr0", adapter);

	return dvrname;
}

static inline char *dmxname(int adapter) {
	static char	dmxname[128];

	sprintf(dmxname, "/dev/dvb/adapter%d/demux0", adapter);

	return dmxname;
}

static int dmx_init(struct adapter_s *adapter) {
	static struct event	dmxevent;
	struct dmx_filter	df;
	int			dmxfd;

	/* FIXME Check for open success */
	dmxfd=open(dmxname(adapter->no), O_RDWR);

	if (dmxfd < 0)
		return 0;

	event_set(&dmxevent, dmxfd, EV_READ|EV_PERSIST, dmx_read, adapter);
	event_add(&dmxevent, NULL);

	memset(&df, 0, sizeof(struct dmx_filter));

	/* Prepare filter - give us section 0x0 aka PAT */
	df.filter[0]=0x0;
	df.mask[0]=0xff;

	/* Set filter and immediatly start receiving packets */
	demux_set_sct_filter(dmxfd, 0, &df, DMX_IMMEDIATE_START|DMX_CHECK_CRC, 0);

	return 1;
}

static int dvr_init(struct adapter_s *adapter) {
	static struct event	pesevent;
	int			dmxfd, dvrfd;

	/* FIXME - check for open success */
	dmxfd=open(dmxname(adapter->no), O_RDWR);
	dvrfd=open(dvrname(adapter->no), O_RDONLY|O_NONBLOCK);

	if (dmxfd < 0 || dvrfd < 0) {
		return 0;
	}

	event_set(&pesevent, dvrfd, EV_READ|EV_PERSIST, dvr_read, adapter);
	event_add(&pesevent, NULL);

	/* Budget Mode - Gimme all you have */
	demux_set_pes_filter(dmxfd, 0x2000, DMX_PES_OTHER);

	return 1;
}

static void pidexpire_init(void );
/*
 * If the PAT or PMT didnt refresh a PID for 30 seconds
 * we remove it from the pid table
 *
 */
static void pidexpire(int fd, short event, void *arg) {
	time_t		now=time(NULL);
	int		i;

	/*
	 * Loop on pids - If the pid expired e.g. nobody renewed
	 * the fwd state expire it and decrease the refcount in
	 * the stream object.
	 *
	 */
#define PID_TIMEOUT	30
	for(i=0;i<MAX_PID;i++) {

		if (pidtable[i].type == PID_NONE)
			continue;

		if (pidtable[i].lastvalid+PID_TIMEOUT < now) {
			printf("expiring pid %d\n", i);

			switch(pidtable[i].type) {
				case(PID_VIDEO):
					pidtable[i].channel->vpid--;
					break;
				case(PID_AUDIO):
					pidtable[i].channel->apid--;
					break;
				case(PID_OTHER):
				case(PID_PMT):
					pidtable[i].channel->opid--;
					break;
			}

			pidtable[i].type = PID_NONE;
		}
	}

	/* Reinit timer */
	pidexpire_init();
}

static void pidexpire_init(void ) {
	static struct event	peevent;
	static struct timeval	tv;

	tv.tv_usec=0;
	tv.tv_sec=30;

	evtimer_set(&peevent, pidexpire, &peevent);
	evtimer_add(&peevent, &tv);
}


int demux_init(struct adapter_s *adapter) {
	/* demux0 PAT receive */
	if (!dmx_init(adapter))
		return 0;

	/* dvr0 stream reveive in budget mode*/
	if (!dvr_init(adapter))
		return 0;

	pidexpire_init();

	return 1;
}


